import logging
import os
import sys
from collections.abc import Callable
from dataclasses import dataclass
from pathlib import Path
from typing import Any, TypeAlias, TypeVar

import tomllib

log = logging.getLogger(__name__)
Secret: TypeAlias = str | dict[str, str] | None

@dataclass
class DomainInfo:
    handler: str
    secret: Secret

def invalid_type_msg(what: str, expected: str, got: Any) -> str:
    return f'invalid type for {what}: expected {expected}, got {type(got).__name__}'

class Config:
    post_acquire: list[str]
    certificates: list[str]
    domains: dict[str, DomainInfo]
    secrets: dict[str, Secret]
    acme_path: Path

    def find_zone(self, domain: str) -> str:
        parts = domain.split('.')
        for i in range(len(parts)-1):
            zone = '.'.join(parts[i:])
            if zone in self.domains:
                return zone
        log.error('could not find zone for domain %s', domain)
        sys.exit(1)

    def get_handler(self, domain: str) -> str:
        if domain in self.domains:
            return self.domains[domain].handler
        raise Exception(f'domain {domain} not found in the config')

    def get_secret(self, domain: str) -> Secret:
        return self.domains[domain].secret

    def get_single_secret(self, domain: str) -> str:
        secret = self.get_secret(domain)
        if type(secret) is not str:
            raise TypeError(f'invalid type for domain "{domain}" secret: expected str, got {type(secret).__name__}')
        return secret

    def get_multi_secret(self, domain: str) -> dict[str, str]:
        secret = self.get_secret(domain)
        if type(secret) is not dict:
            raise TypeError(f'invalid type for domain "{domain}" secret: expected dict, got {type(secret).__name__}')
        return secret

def config_parse_secret(raw: Any) -> Secret:
    if type(raw) is str:
        return raw
    if type(raw) is dict:
        for k, v in raw:
            if type(k) is not str:
                raise TypeError(f'invalid type for secret name: expected str, got {type(k).__name__}')
            if type(v) is not str:
                raise TypeError(f'invalid type for secret "{k}" value: expected str, got {type(v).__name__}')
        return raw
    raise TypeError(f'invalid type for secret: expected str or dict, got {type(raw).__name__}')

def config_parse_domain(raw: Any) -> DomainInfo:
    if type(raw) is str:
        return DomainInfo(handler=raw, secret=None)
    if type(raw) is dict:
        if 'handler' not in raw:
            raise TypeError('domain info does not contain "handler"')
        handler = raw['handler']
        if type(handler) is not str:
            raise TypeError(f'invalid type for domain info handler: expected str, got {type(handler).__name__}')
        secret = None
        if 'secret' in raw:
            secret = config_parse_secret(raw['secret'])
        return DomainInfo(handler=handler, secret=secret)
    raise TypeError(f'invalid type for domain info: expected str or dict, got {type(raw).__name__}')

T = TypeVar('T')

def config_parse_dict(raw_conf: dict[str, Any], key: str, parse_item: Callable[[Any], T]) -> dict[str, T]:
    if key not in raw_conf:
        log.error(f'missing "{key}" in config')
        sys.exit(1)

    result: dict[str, T] = {}

    for k, v in raw_conf[key].items():
        if not isinstance(k, str):
            raise TypeError(f'"{k}" is not a string')
        result[k] = parse_item(v)

    return result

def config_parse_list(raw_conf: dict[str, Any], key: str) -> list[str]:
    result: list[str] = []

    if not isinstance(raw_conf[key], list):
        raise TypeError(f'invalid type for "{key}": expected list, not {type(raw_conf[key]).__name__}')

    for item in raw_conf[key]:
        if type(item) is not str:
            raise TypeError(f'invalid type for "{key}" list item: expected str, got {type(item).__name__}')
        result.append(item)

    return result


def read_config(path: str | None) -> Config:
    if not path:
        # should be here only when running from hook
        path = os.getenv('NYACME_CONFIG', '/etc/nyacme.toml')

    with open(path, 'rb') as file:
        raw_conf = tomllib.load(file)

    for key in raw_conf:
        if key not in ('domains', 'secrets', 'post_acquire', 'acme_path', 'certificates'):
            log.warning('unknown config key: %s', key)

    c = Config()

    c.domains = config_parse_dict(raw_conf, 'domains', config_parse_domain)
    secrets = config_parse_dict(raw_conf, 'secrets', config_parse_secret)
    # fill default secrets
    for domain in c.domains:
        info = c.domains[domain]
        if info.secret is None:
            if info.handler not in secrets:
                raise KeyError(f'"{info.handler}" not in secrets')
            info.secret = secrets[info.handler]

    post_acquire = []
    if 'post_acquire' in raw_conf:
        post_acquire = config_parse_list(raw_conf, 'post_acquire')

    c.post_acquire = post_acquire

    if 'acme_path' in raw_conf:
        c.acme_path = Path(raw_conf['acme_path'])
    else:
        c.acme_path = Path('/var/www/acme/.well-known/acme-challenge')

    if 'certificates' not in raw_conf:
        log.error('missing "certificates"')
        sys.exit(1)

    c.certificates = config_parse_list(raw_conf, 'certificates')

    return c
