import json
import os
import subprocess
from typing import Any

import dns.resolver

from ..config import Config
from .base import Handler


class CloudflareHandler(Handler):
    nameservers: list[str]

    def __init__(self, zone_name: str, config: Config, token: str) -> None:
        super().__init__(zone_name, config, token)
        self.secret = config.get_single_secret(zone_name)

        servers = dns.resolver.resolve(zone_name, 'NS')
        self.nameservers = [ str(rdata.target).strip('.') for rdata in servers ]

    def run(self, args: list[str]) -> Any:
        env = os.environ.copy()
        env['CF_API_TOKEN'] = self.secret
        res = subprocess.run(['flarectl', '--json', *args], check=True, stdout=subprocess.PIPE, env=env)
        try:
            return json.loads(res.stdout)
        except json.decoder.JSONDecodeError as ex:
            self.log.warning('could not decode as JSON: %s', res.stdout)
            raise ex

    def create(self, record_name: str, record_value: str) -> None:
        full_record_name = record_name + '.' + self.zone
        self.remove(record_name)
        self.log.info('creating %s with value %s', full_record_name, record_value)
        self.run([
            'dns', 'create',
            '--zone', self.zone,
            '--type', 'TXT',
            '--name', full_record_name,
            '--content', record_value,
        ])

    def remove(self, record_name: str) -> None:
        records = self.run(['dns', 'list', '--zone', self.zone])
        full_record_name = record_name + '.' + self.zone
        for record in records:
            if record['Name'] == full_record_name:
                self.log.info('removing %s', full_record_name)
                self.run([
                    'dns', 'delete',
                    '--zone', self.zone,
                    '--id', record['ID'],
                ])
