import json
import urllib.request
from typing import Any

from ..config import Config
from .base import Handler


class PorkbunHandler(Handler):
    # discovered
    nameservers: list[str]

    def __init__(self, zone_name: str, config: Config, token: str) -> None:
        super().__init__(zone_name, config, token)
        secret = config.get_multi_secret(zone_name)
        if 'apikey' not in secret:
            raise TypeError('"apikey" missing')
        if 'secretapikey' not in secret:
            raise TypeError('"secretapikey" missing')
        self.apikey = secret['apikey']
        self.secretapikey = secret['secretapikey']

        self.nameservers = self.fetch(f'/domain/getNs/{self.zone}')['ns']

    def fetch(self, url: str, data: dict[str, Any] | None = None) -> Any:
        req = urllib.request.Request('https://api.porkbun.com/api/json/v3' + url)

        if data is None:
            data = {}
        data['apikey'] = self.apikey
        data['secretapikey'] = self.secretapikey
        req.data = json.dumps(data).encode('utf-8')
        req.add_header('Content-Type', 'application/json;charset=utf-8')
        req.method = 'POST'
        try:
            with urllib.request.urlopen(req) as f:
                res = json.load(f)
                if res['status'] != 'SUCCESS':
                    raise Exception(res['message'])
                return res
        except urllib.error.HTTPError as ex:
            self.log.error('cannot %s %s: %s', req.method, url, ex)
            res = ex.fp.read().decode('utf-8')
            try:
                raise Exception(json.loads(res)['message'])
            except json.JSONDecodeError:
                raise Exception(res) from None

    def create(self, record_name: str, record_value: str) -> None:
        self.remove(record_name)
        self.log.info('creating %s with value %s', record_name, record_value)
        self.fetch(f'/dns/create/{self.zone}', {
            'name': record_name,
            'type': 'TXT',
            'content': record_value,
            'ttl': 300,
        })

    def remove(self, record_name: str) -> None:
        full_record_name = record_name + '.' + self.zone
        records = self.fetch(f'/dns/retrieve/{self.zone}')['records']
        for record in records:
            if record['name'] == full_record_name:
                self.log.info('removing %s', full_record_name)
                self.fetch(f'/dns/delete/{self.zone}/{record['id']}')
